@@ -81,6 +81,17 @@ def sort_packages(
             + (package.exec_depends if test_depends else [])
             + (package.test_depends if test_depends else [])
         )
+        tweaks = get_build_var(package.name, "ROS_DEPENDS", "", environ).split()
+        if test_depends:
+            tweaks += get_build_var(
+                package.name, "ROS_TEST_DEPENDS", "", environ
+            ).split()
+        tweaks = [t for t in tweaks if t]
+        for tweak in tweaks:
+            if tweak[0] == "+":
+                depends.add(tweak[1:])
+            elif tweak[0] == "-":
+                depends.discard(tweak[1:])
         graph[package.name] = list(depends)
     sorted_names = topological_sort(graph)
     return [lookup[name] for name in sorted_names]
@@ -128,7 +139,7 @@ def compute_ament_cmake_python_path(builddir: Path) -> List[str]:
     return []
 
 
-def compute_python_package_path(prefix: Path) -> List[str]:
+def get_python_versions():
     global SUPPORTED_PYTHON_VERSIONS, DEFAULT_PYTHON_VERSION
 
     if DEFAULT_PYTHON_VERSION is None or SUPPORTED_PYTHON_VERSIONS is None:
@@ -137,18 +148,28 @@ def compute_python_package_path(prefix: Path) -> List[str]:
         except ValueError:
             major = 3
         DEFAULT_PYTHON_VERSION, SUPPORTED_PYTHON_VERSIONS = cpython_versions(major)
+    return DEFAULT_PYTHON_VERSION, SUPPORTED_PYTHON_VERSIONS
+
+
+def is_default_python(version) -> bool:
+    default_python, _ = get_python_versions()
+    return version == default_python
+
+
+def compute_python_package_path(prefix: Path) -> List[str]:
+    default_python, supported_python = get_python_versions()
 
     def append_sitedir(path):
         if path.is_dir():
             result.append(str(path))
 
     result = []
-    if DEFAULT_PYTHON_VERSION is not None:
-        if DEFAULT_PYTHON_VERSION[0] == 3:
+    if default_python is not None:
+        if default_python[0] == 3:
             append_sitedir(prefix / "lib" / "python3" / "dist-packages")
 
-    if SUPPORTED_PYTHON_VERSIONS is not None:
-        for version in SUPPORTED_PYTHON_VERSIONS:
+    if supported_python is not None:
+        for version in supported_python:
             append_sitedir(
                 prefix / "lib" / "python{}.{}".format(*version[:2]) / "dist-packages"
             )
