@@ -16,6 +16,8 @@ from ..common import (
     compute_ament_cmake_python_path,
     sort_packages,
     get_build_var,
+    get_python_versions,
+    is_default_python,
     query_dpkg_architecture,
 )
 
@@ -118,6 +120,15 @@ def run(args, environ):
         skipped_tests = set(p[1].name for p in packages)
     else:
         skipped_tests = set(shlex.split(skipped_tests) if skipped_tests else [])
+    python_extension_packages = get_build_var(
+        None, "ROS_PYTHON_EXTENSION_PACKAGES", env=environ
+    )
+    if python_extension_packages == "1":
+        python_extension_packages = set(p[1].name for p in packages)
+    else:
+        python_extension_packages = set(
+            shlex.split(python_extension_packages) if python_extension_packages else []
+        )
     ignore_test_results = get_build_var(None, "ROS_IGNORE_TEST_RESULTS", env=environ)
     if ignore_test_results == "1":
         ignore_test_results = set(p[1].name for p in packages)
@@ -138,7 +149,14 @@ def run(args, environ):
             for path, package in packages:
                 if package.name not in skipped_packages:
                     do_action(
-                        "clean", args, path, package, buildopts, install_prefix, environ
+                        "clean",
+                        args,
+                        path,
+                        package,
+                        buildopts,
+                        install_prefix,
+                        environ,
+                        None,
                     )
             if args.builddir.is_dir():
                 log.info(f"Removing {args.builddir}")
@@ -167,7 +185,7 @@ def run(args, environ):
         destdirs = set()
         for path, package in packages:
             if package.name in skipped_packages:
-                log.info("Skipping ROS package {package.name}")
+                log.info(f"Skipping ROS package {package.name}")
                 continue
 
             sys.stdout.write(
@@ -176,34 +194,83 @@ def run(args, environ):
                 f"= ROS Package {package.name:<61} =\n"
                 "=============================================================================\n"
             )
+            if package.get_build_type() == "catkin":
+                catkin_srcs.append(str(path.resolve()))
+            python_versions = [None]
             package_buildopts = []
             if package.name in skipped_tests:
                 package_buildopts.append("nocheck")
-            do_action(
-                "configure",
-                args,
-                path,
-                package,
-                buildopts + package_buildopts,
-                install_prefix,
-                environ,
-            )
-            do_action("build", args, path, package, buildopts, install_prefix, environ)
-            if "nocheck" not in buildopts and package.name not in skipped_tests:
-                try:
-                    do_action(
-                        "test", args, path, package, buildopts, install_prefix, environ
-                    )
-                except subprocess.CalledProcessError as e:
-                    if package.name in ignore_test_results:
-                        log.info(f"Ignoring test failure in ROS package {package.name}")
-                    else:
-                        raise
-            else:
-                log.info(f"Skipping tests for ROS package {package.name}")
-            do_action(
-                "install", args, path, package, buildopts, install_prefix, environ
-            )
+            if (
+                package.name in python_extension_packages
+                and package.get_build_type()
+                in [
+                    "cmake",
+                    "catkin",
+                    "ament_cmake",
+                ]
+            ):
+                default_python, supported_python = get_python_versions()
+                python_versions = [
+                    tuple(v) for v in supported_python if v != default_python
+                ]
+                python_versions.append(tuple(default_python))
+                s = [".".join(str(v) for v in pv) for pv in python_versions]
+                log.info(
+                    f"Python extensions will be built against Python {', '.join(s)}"
+                )
+
+            for python_version in python_versions:
+                do_action(
+                    "configure",
+                    args,
+                    path,
+                    package,
+                    buildopts + package_buildopts,
+                    install_prefix,
+                    environ,
+                    python_version,
+                )
+                do_action(
+                    "build",
+                    args,
+                    path,
+                    package,
+                    buildopts,
+                    install_prefix,
+                    environ,
+                    python_version,
+                )
+                if "nocheck" not in buildopts and package.name not in skipped_tests:
+                    try:
+                        do_action(
+                            "test",
+                            args,
+                            path,
+                            package,
+                            buildopts,
+                            install_prefix,
+                            environ,
+                            python_version,
+                        )
+                    except subprocess.CalledProcessError as e:
+                        if package.name in ignore_test_results:
+                            log.info(
+                                f"Ignoring test failure in ROS package {package.name}"
+                            )
+                        else:
+                            raise
+                else:
+                    log.info(f"Skipping tests for ROS package {package.name}")
+                do_action(
+                    "install",
+                    args,
+                    path,
+                    package,
+                    buildopts,
+                    install_prefix,
+                    environ,
+                    python_version,
+                )
         sys.stdout.write(
             "\n"
             "#############################################################################\n"
@@ -218,7 +285,9 @@ def run(args, environ):
     return 0
 
 
-def do_action(action, args, path, package, buildopts, install_prefix, environ):
+def do_action(
+    action, args, path, package, buildopts, install_prefix, environ, python_version
+):
     global destdirs
     global catkin_srcs
     log = logging.getLogger("dhros")
@@ -228,23 +297,31 @@ def do_action(action, args, path, package, buildopts, install_prefix, environ):
     homedir = Path(
         get_build_var(package.name, "ROS_HOME", args.builddir / ".roshome", env=environ)
     ).resolve()
+    builddir = args.builddir / package.name
+    python_interpreter = ""
+    python_version_str = ""
+    if python_version is not None:
+        python_version_str = ".".join(str(v) for v in python_version)
+        builddir = builddir / f"py{python_version_str}"
+        if is_default_python(python_version):
+            python_interpreter = f"/usr/bin/python{python_version[0]}"
+        else:
+            python_interpreter = f"/usr/bin/python{python_version_str}"
     catkin_test_results_dir = Path(
         get_build_var(
             package.name,
             "ROS_CATKIN_TEST_RESULTS_DIR",
-            args.builddir / package.name / "test_results",
+            builddir / "test_results",
             env=environ,
         )
     ).resolve()
-    catkin_devel_dir = (args.builddir / "__devel__").resolve()
+    catkin_devel_dir = (args.builddir / "_devel").resolve()
     relative_install_prefix = install_prefix.relative_to("/")
     destdirs.add(destdir)
     package_suffix = re.sub(r"[^A-Za-z0-9]", "_", package.name)
     build_type = package.get_build_type()
-    if build_type == "catkin":
-        catkin_srcs.append(str(path.resolve()))
     hook_vars = {
-        "build_dir": shlex.quote(str((args.builddir / package.name).resolve())),
+        "build_dir": shlex.quote(str(builddir.resolve())),
         "build_type": shlex.quote(build_type),
         "destdir": shlex.quote(str(destdir)),
         "dir": shlex.quote(str(path.resolve())),
@@ -253,6 +330,8 @@ def do_action(action, args, path, package, buildopts, install_prefix, environ):
         "package!id": package_suffix,
         "prefix": shlex.quote(str(install_prefix)),
         "version": shlex.quote(package.version),
+        "python_version": python_version_str,
+        "python_interpreter": python_interpreter,
     }
     env = environ.copy()
     pybuild_per_package_keys = [
@@ -284,8 +363,6 @@ def do_action(action, args, path, package, buildopts, install_prefix, environ):
     prefixes = prefixes.split(":") if prefixes is not None else []
     if install_prefix != Path("/usr"):
         prefixes.append(str(install_prefix))
-    if build_type == "catkin":
-        prefixes.append(str(catkin_devel_dir))
     env["CMAKE_PREFIX_PATH"] = ":".join(
         prefixes + [str(d / relative_install_prefix) for d in destdirs]
     )
@@ -311,7 +388,7 @@ def do_action(action, args, path, package, buildopts, install_prefix, environ):
     pythonpaths = env.get("PYTHONPATH", None)
     pythonpaths = pythonpaths.split(":") if pythonpaths is not None else []
     if build_type == "ament_cmake":
-        pythonpaths += compute_ament_cmake_python_path(args.builddir / package.name)
+        pythonpaths += compute_ament_cmake_python_path(builddir)
     if build_type == "catkin":
         pythonpaths += compute_python_package_path(catkin_devel_dir)
     for d in destdirs:
@@ -363,7 +440,7 @@ def do_action(action, args, path, package, buildopts, install_prefix, environ):
             cmdline += ["--buildsystem=pybuild"]
         elif build_type in ["ament_cmake", "catkin", "cmake"]:
             cmdline += [
-                f"--builddir={args.builddir / package.name}",
+                f"--builddir={builddir}",
                 "--buildsystem=cmake",
             ]
         if action == "install":
@@ -394,8 +471,20 @@ def do_action(action, args, path, package, buildopts, install_prefix, environ):
                         "-DAMENT_CMAKE_ENVIRONMENT_PACKAGE_GENERATION=OFF",
                         "-DAMENT_CMAKE_ENVIRONMENT_PARENT_PREFIX_PATH_GENERATION=OFF",
                     ]
-        elif action == "test" and build_type == "catkin" and package.name != "catkin":
-            cmdline += ["run_tests"]
+            if python_version is not None:
+                cmdline += [
+                    f"-DPYTHON_EXECUTABLE={python_interpreter}",
+                    f"-DPython_EXECUTABLE={python_interpreter}",
+                    f"-DPython{python_version[0]}_EXECUTABLE={python_interpreter}",
+                ]
+        elif action == "test" and build_type == "catkin":
+            cmdline = [
+                "cmake",
+                "--build",
+                str(builddir),
+                "--target",
+                "run_tests",
+            ]
 
         for extra_args_var in [
             f"ROS_{action.upper()}_{build_type.upper()}_WRAPPER",
@@ -403,7 +492,13 @@ def do_action(action, args, path, package, buildopts, install_prefix, environ):
         ]:
             extra_args = get_build_var(package.name, extra_args_var, env=environ)
             if extra_args:
-                cmdline = shlex.split(extra_args.format(hook_vars)) + cmdline
+                try:
+                    cmdline = shlex.split(extra_args.format(**hook_vars)) + cmdline
+                except KeyError as e:
+                    log.error(
+                        f"Invalid substitution {{{str(e)}}} in {extra_args_var}: {extra_args!r}"
+                    )
+                    raise
                 break
         for extra_args_var in [
             f"ROS_{action.upper()}_ARGS",
@@ -411,7 +506,13 @@ def do_action(action, args, path, package, buildopts, install_prefix, environ):
         ]:
             extra_args = get_build_var(package.name, extra_args_var, env=environ)
             if extra_args:
-                cmdline += shlex.split(extra_args.format(hook_vars))
+                try:
+                    cmdline += shlex.split(extra_args.format(**hook_vars))
+                except KeyError as e:
+                    log.error(
+                        f"Invalid substitution {{{str(e)}}} in {extra_args_var}: {extra_args!r}"
+                    )
+                    raise
 
         sys.stdout.write(cmdind)
         sys.stdout.write(shlex.join(cmdline))
@@ -459,7 +560,7 @@ def execute_hook(hook, package, vars, env, dry_run=False):
         return False
 
     try:
-        cmdline = cmdline.format(vars)
+        cmdline = cmdline.format(**vars)
     except KeyError as e:
         log.error(f"Invalid substitution {{{str(e)}}} in hook {hook!r}")
         raise
